/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <stdio.h>
#include <getopt.h>
#include <signal.h>
#include "tcpjunk.h"
#include "client.h"
#include "server.h"
#include "gui.h"
#include "utils.h"
#include "net.h"

int main(int argc, char *argv[])
  {
  /* ignore sigpipe */
  signal(SIGPIPE, SIG_IGN);

  /* clean dead processes */
  struct sigaction sig_a;
  sig_a.sa_handler = sigchld_handler;
  sigemptyset(&sig_a.sa_mask);
  sig_a.sa_flags = SA_RESTART;
  sigaction(SIGCHLD, &sig_a, NULL);

  /* init instance data */
  struct tj_data inst_data;
  inst_data.debug = 0;
  inst_data.verbose = 0;
  inst_data.hexdump = 0;
  inst_data.gui = 0;
  inst_data.wait = 0;
  inst_data.conn_timeout = 2;
  inst_data.reply_timeout = 100;
  inst_data.reply_timeout *= 1000;
  inst_data.keepalive = 1;
  inst_data.loop = 0;
  inst_data.wait_conn = 0;
  inst_data.no_wait_last = 0;
  inst_data.use_ssl = 0;
  inst_data.server_cert[0] = '\0';
  inst_data.server_key[0] = '\0';
  inst_data.port[0] = '\0';
  inst_data.src_port[0] = '\0';
  inst_data.interface[0] = '\0';
  inst_data.src_ip[0] = '\0';
  inst_data.sess_file_name[0] = '\0';
  inst_data.offset = 0;
  inst_data.send_delay = 0;
  inst_data.send_size = 0;
  inst_data.max_seg = 0;
  inst_data.inst_num = 1;
  inst_data.auto_file = 0;
  inst_data.server_mode = 0;
  inst_data.sess_data = NULL;
  inst_data.sess_data = 0;
  inst_data.ipv6 = 0;
  inst_data.remote_addr = NULL;
  inst_data.local_addr = NULL;
  inst_data.src_ip_num_ses = 10;

  int insts_num = 0;
  const char *rhost_addr_str = NULL;
  const char *auto_file_dir = NULL;
  int tcp_opt = 0;
  int do_daemonize = 0;

  /* parse arguments */
  int next_option;
  const char* const short_options =
	  "ahxvrs:p:c:f:e:i:dw:t:T:WCN:6o:D:m:zXk:l:SE:K:P:I:O:M:";

  const struct option long_options[] =
	{
	  {
	  "help", 0, NULL, 'h'
	  },
	  {
	  "debug", 0, NULL, 'd'
	  },
	  {
	  "wait-data", 0, NULL, 'W'
	  },
	  {
	  "optimize", 0, NULL, 'z'
	  },
	  {
	  "close", 0, NULL, 'C'
	  },
	  {
	  "gui", 0, NULL, 'x'
	  },
	  {
	  "verbose", 0, NULL, 'v'
	  },
	  {
	  "daemon", 0, NULL, 'a'
	  },
	  {
	  "ipv6", 0, NULL, '6'
	  },
	  {
	  "hexdump", 0, NULL, 'X'
	  },
	  {
	  "server", 0, NULL, 'r'
	  },
	  {
	  "instances", 1, NULL, 'i'
	  },
	  {
	  "host", 1, NULL, 's'
	  },
	  {
	  "port", 1, NULL, 'p'
	  },
	  {
	  "src-port", 1, NULL, 'o'
	  },
	  {
	  "ip-sessions", 1, NULL, 'N'
	  },
	  {
	  "timeout", 1, NULL, 't'
	  },
	  {
	  "offset", 1, NULL, 'O'
	  },
	  {
	  "interface", 1, NULL, 'I'
	  },
	  {
	  "src-ip", 1, NULL, 'P'
	  },
	  {
	  "sess-timeout", 1, NULL, 'T'
	  },
	  {
	  "send-size", 1, NULL, 'm'
	  },
	  {
	  "file", 1, NULL, 'c'
	  },
	  {
	  "send-delay", 1, NULL, 'D'
	  },
	  {
	  "repeat-sess", 1, NULL, 'k'
	  },
	  {
	  "limit-sess", 1, NULL, 'l'
	  },
	  {
	  "rand-file", 1, NULL, 'f'
	  },
	  {
	  "seq-file", 1, NULL, 'e'
	  },
	  {
	  "wait", 1, NULL, 'w'
	  },
	  {
	  "mss", 1, NULL, 'M'
	  },
	  {
	  "ssl", 0, NULL, 'S'
	  },
	  {
	  "cert", 1, NULL, 'E'
	  },
	  {
	  "key", 1, NULL, 'K'
	  },
	  {
	  NULL, 0, NULL, 0
	  }
	};

  do
	{
	next_option = getopt_long(argc, argv, short_options, long_options, NULL);
	switch (next_option)
	  {
	/* help */
	case 'h':
	  tj_show_help();
	  exit(EXIT_SUCCESS);

	  /* GUI */
	case 'x':
	  gui_init(argc, argv, &inst_data);
	  exit(EXIT_SUCCESS);

	  /* verbose */
	case 'v':
	  inst_data.verbose = 2;
	  break;

	  /* daemonize */
	case 'a':
	  do_daemonize = 1;
	  break;

	  /* optimize */
	case 'z':
	  tcp_opt = 1;
	  break;

	  /* wait for data */
	case 'W':
	  inst_data.wait_conn = 1;
	  break;

	  /* close */
	case 'C':
	  inst_data.no_wait_last = 1;
	  break;

	  /* max_seg */
	case 'M':
	  inst_data.max_seg = atoi(optarg);
	  break;

	  /* debug */
	case 'd':
	  inst_data.debug = 2;
	  break;

	  /* ipv6 */
	case '6':
	  inst_data.ipv6 = 1;
	  break;

	  /* hexdump */
	case 'X':
	  inst_data.hexdump = 1;
	  break;

	  /* server mode */
	case 'r':
	  inst_data.server_mode = 1;
	  break;

	  /* host */
	case 's':
	  rhost_addr_str = optarg;
	  break;

	  /* number of instances */
	case 'i':
	  insts_num = atoi(optarg);
	  break;

	  /* wait between connections */
	case 'w':
	  inst_data.wait = atoi(optarg) * 1000;
	  break;

	  /* repeat session */
	case 'k':
	  inst_data.keepalive = atoi(optarg);
	  break;

	  /* limit session */
	case 'l':
	  inst_data.loop = atoi(optarg);
	  break;

	  /* send size */
	case 'm':
	  inst_data.send_size = atoi(optarg);
	  break;

	  /* send delay */
	case 'D':
	  inst_data.send_delay = atoi(optarg) * 1000;
	  break;

	  /* number of sessions for source IP */
	case 'N':
	  inst_data.src_ip_num_ses = atoi(optarg);
	  break;

	  /* connection timeout */
	case 't':
	  inst_data.conn_timeout = atoi(optarg);
	  break;

	  /* session timeout */
	case 'T':
	  inst_data.reply_timeout = atoi(optarg) * 1000;
	  break;

	  /* alias offset */
	case 'O':
	  inst_data.offset = atoi(optarg) * 1000;
	  break;

	  /* port */
	case 'p':
	  strcpy(inst_data.port, optarg);
	  break;

	  /* source port */
	case 'o':
	  strcpy(inst_data.src_port, optarg);
	  break;

	  /* session file */
	case 'c':
	  strcpy(inst_data.sess_file_name, optarg);
	  break;

	  /* random session file */
	case 'f':
	  auto_file_dir = optarg;
	  inst_data.auto_file = f_rand;
	  break;

	  /* sequential session file */
	case 'e':
	  auto_file_dir = optarg;
	  inst_data.auto_file = f_seq;
	  break;

	  /* source IP */
	case 'P':
	  strcpy(inst_data.src_ip, optarg);
	  break;

	  /* interface */
	case 'I':
	  strcpy(inst_data.interface, optarg);
	  break;

	  /* SSL */
	case 'S':
	  inst_data.use_ssl = 1;
	  break;

	  /* SSL certificate */
	case 'E':
	  strcpy(inst_data.server_cert, optarg);
	  break;

	  /* SSL key */
	case 'K':
	  strcpy(inst_data.server_key, optarg);
	  break;

	case '?':
	  exit(EXIT_FAILURE);

	case -1:
	  break;

	default:
	  exit(EXIT_FAILURE);
	  }
	}
  while (next_option != -1);

  /* check client mode arguments */
  if (inst_data.server_mode == 0)
	{
	/* remote host */
	if (!rhost_addr_str)
	  {
	  printf("tcpjunk: remote host not specified\r\n");
	  exit(EXIT_FAILURE);
	  }

	/* session limit */
	if (inst_data.loop > 0 && insts_num == 0)
	  {
	  printf(
		  "tcpjunk: session limit only relevant when using multiple instances\r\n");
	  exit(EXIT_FAILURE);
	  }

	/* number of instances and debug */
	if (insts_num == 0)
	  {
	  insts_num = 1;
	  inst_data.debug = 1;
	  inst_data.verbose = 1;
	  }
	else
	  {
	  /* if no force debug and verbose - turn it off */
	  if (inst_data.debug != 2)
		inst_data.debug = 0;
	  if (inst_data.verbose != 2)
		inst_data.verbose = 0;

	  if (insts_num < 0 || insts_num > 1000)
		{
		printf("tcpjunk: invalid number of instances \'%d\'\r\n", insts_num);
		exit(EXIT_FAILURE);
		}
	  }
	/* source IP */
	if (strcmp(inst_data.interface, ""))
	  {
	  if (!inst_data.src_ip[0])
		{
		printf("tcpjunk: source IP must be specified for -I option\n");
		exit(EXIT_FAILURE);
		}
	  }
	}

  /* check port */
  if (!inst_data.port[0])
	{
	printf("tcpjunk: port not specified\r\n");
	exit(EXIT_FAILURE);
	}

  /* check destination address */
  if (inst_data.server_mode == 0)
	inst_data.remote_addr = check_address(&inst_data, rhost_addr_str,
		inst_data.port, ai_none);

  /* check file or load file list */
  if (inst_data.auto_file)
	load_file_list(&inst_data, auto_file_dir);
  else if (!inst_data.sess_file_name[0])
	{
	printf("tcpjunk: session file not specified\r\n");
	exit(EXIT_FAILURE);
	}
  else
	{
	/* load session file */
	load_session_data(&inst_data);

	/* check session data */
	find_sections(&inst_data);
	}

  /* optimize TCP */
  if (tcp_opt == 1)
	{
	int sys_rv = 0;
	sys_rv = system("ulimit -n 10240");
	sys_rv += system("echo 1 > /proc/sys/net/ipv4/tcp_tw_recycle");
	sys_rv += system("echo 1 > /proc/sys/net/ipv4/tcp_tw_reuse");
	if (sys_rv)
	  {
	  printf("tcpjunk: TCP parameters optimization failed\n");
	  exit(EXIT_FAILURE);
	  }
	}

  /* daemonize */
  if (do_daemonize)
	daemonize();

  /* start client or server */
  if (inst_data.server_mode == 0)
	client_init(&inst_data, insts_num);
  else
	server_init(&inst_data);

  return EXIT_SUCCESS;
  }

